;; ***************************************************************************
;; funcio copiar_tauler: Fa una copia d'un tauler de 4x4
;; ***************************************************************************

(defun copiar_tauler (tauler)
  (make-array '(4 4) :initial-contents (valors tauler 0))
)

;; Valors: copia els valors del tauler des de la fila donada
;; Com es veu a copiar_tauler, la primera fila es la 0

(defun valors (tau f)
  (if (= f 4)
    nil
    (cons (fila tau f 0)  (valors tau (+ f 1)))
  )
)

;;; Fila: copia tots els valors d'una fila, des de la columna donada
;;; Com es veu a valors, la primera columna es la 0

(defun fila (tau f c)
  (if (= c 4)
    nil                                         
    (cons (aref tau f c)  (fila tau f  (+ c 1)))      
  )
)
;; ***************************************************************************



; Afegeix una ficha a la posici indicada del tauler
; node es array 4x4, lloc es la columna enumerada de 0 a 3 (esquerra a dreta)
; i tipus es 1 per les X (pc), 2 per les O (oponent)
(defun ficha(node lloc tipus)
  (if (= (aref node 3 lloc) 0)
    (setf (aref node 3 lloc) tipus)
    (if (= (aref node 2 lloc) 0)
      (setf (aref node 2 lloc) tipus)
      (if (= (aref node 1 lloc) 0)
        (setf (aref node 1 lloc) tipus)
        (if (= (aref node 0 lloc) 0)
          (setf (aref node 0 lloc) tipus)
          node  ; No es pot afegir ms
        )
      )
    )
  )
  node
)


; Retorna en quina posici es troba el maxim
(defun maxpos (llista)
  (maxpos_rec llista -2000 0 0)
)

; Inicialment max ha de ser -2000 (- infinit), pos 0 i total 0
(defun maxpos_rec (llista max pos total)
  (if (null llista)
    pos   ; Retornem el maxim
    (if (< max (car llista))
      (maxpos_rec (cdr llista) (car llista) total (+ total 1))
      (maxpos_rec (cdr llista) max pos (+ total 1))
    )
  )
)

; Comproba si esta ple el tauler
(defun ple (node)
  (if (> (aref node 0 0) 0)
    (if (> (aref node 0 1) 0)
      (if (> (aref node 0 2) 0)
        (if (> (aref node 0 3) 0)
          T
          nil
        )
        nil
      )
      nil
    )
    nil
  )
)

; Comprova si ha guanyat un jugador al tauler
; Puntua indica quin jugador estem comprovant si ha guanyat:
;  1 -> Ordinador
;  2 -> Usuari
(defun guanya (node puntua)
  (if (guanya_fila node 0 0 puntua)
    T
    (if (guanya_columna node 0 0 puntua)
      T
      (if (guanya_diagonal1 node 3 2 puntua)
        T
        (if (guanya_diagonal1 node 3 3 puntua)
          T
          (if (guanya_diagonal1 node 2 2 puntua)
            T
            (if (guanya_diagonal1 node 2 3 puntua)
              T
              (if (guanya_diagonal2 node 3 1 puntua)
                T
                (if (guanya_diagonal2 node 3 0 puntua)
                  T
                  (if (guanya_diagonal2 node 2 1 puntua)
                    T
                    (if (guanya_diagonal2 node 2 0 puntua)
                      T
                      nil
                    )
                  )
                )
              )
            )
          )
        )
      )
    )
  )
)


; Busca per totes les fileres una combinaci del tipus:
;  111 si puntua val 1 (guanya ordinador)
;  222 si puntua val 2 (guanya usuari)
(defun guanya_fila (node y x puntua)
  (if (= y 4)
    nil
    (if (= (aref node y x) puntua)
      (if (= (aref node y (+ x 1)) puntua) ;1---
        (if (= (aref node y (+ x 2)) puntua) ;11--
          T
          (guanya_fila node (+ y 1) x puntua)
        )
        (guanya_fila node (+ y 1) x puntua)
      )
      (if (= (aref node y (+ x 1)) puntua) ;0--
        (if (= (aref node y (+ x 2)) puntua) ;01--
          (if (= (aref node y (+ x 3)) puntua) ;011-
            T
            (guanya_fila node (+ y 1) x puntua)
          )
          (guanya_fila node (+ y 1) x puntua)
        )
        (guanya_fila node (+ y 1) x puntua)
      )
    )
  )
)

; Busca per totes les columnes una combinaci del tipus:
;  111 si puntua val 1 (guanya ordinador)
;  222 si puntua val 2 (guanya usuari)
(defun guanya_columna (node y x puntua)
  (if (= x 4)
    nil
    (if (= (aref node y x) puntua)
      (if (= (aref node (+ y 1) x) puntua) ;1---
        (if (= (aref node (+ y 2) x) puntua) ;11--
          T
          (guanya_columna node y (+ x 1) puntua)
        )
        (guanya_columna node y (+ x 1) puntua)
      )
      (if (= (aref node (+ y 1) x) puntua) ;0--
        (if (= (aref node (+ y 2) x) puntua) ;01--
          (if (= (aref node (+ y 3) x) puntua) ;011-
            T
            (guanya_columna node y (+ x 1) puntua)
          )
          (guanya_columna node y (+ x 1) puntua)
        )
        (guanya_columna node y (+ x 1) puntua)
      )
    )
  )
)

; Busca per totes les diagonals amb sentit baix dreta cap a amunt esquerra
; una combinaci del tipus:
;  111 si puntua val 1 (guanya ordinador)
;  222 si puntua val 2 (guanya usuari)
(defun guanya_diagonal1 (node y x puntua)
      (if (= (aref node y x) puntua)
        (if (= (aref node (- y 1) (- x 1)) puntua)
          (if (= (aref node (- y 2) (- x 2)) puntua)
            T ; 111
            nil
          )
          nil
        )
        nil
      )
)

; Busca per totes les diagonals amb sentit baix esquerra cap a amunt dreta
; una combinaci del tipus:
;  111 si puntua val 1 (guanya ordinador)
;  222 si puntua val 2 (guanya usuari)
(defun guanya_diagonal2 (node y x puntua)
      (if (= (aref node y x) puntua)
        (if (= (aref node (- y 1) (+ x 1)) puntua)
          (if (= (aref node (- y 2) (+ x 2)) puntua)
            T ; 111
            nil
          )
          nil
        )
        nil
      )
)

; Funci heurstica
; Si guanya l'ordinador retorna 1000, si perd retorna -1000 i sino
; realitza un sumatori de l'estat del tauler que sempre estar entre [-1000, 1000]
(defun h (node)
  (if (guanya node 1)
    1000
    (if (guanya node 2)
      -1000
      (- 
        (+ (h_y node 0 0 1) (h_y node 0 1 1) (h_x node 0 0 1) (h_x node 1 0 1)  (h_diagonal1 node 3 2 1) (h_diagonal1 node 3 3 1) (h_diagonal1 node 2 2 1) (h_diagonal1 node 2 3 1) (h_diagonal2 node 3 1 1) (h_diagonal2 node 3 0 1) (h_diagonal2 node 2 1 1) (h_diagonal2 node 2 0 1))
        (+ (h_y node 0 0 2) (h_y node 0 1 2) (h_x node 0 0 2) (h_x node 1 0 2)  (h_diagonal1 node 3 2 2) (h_diagonal1 node 3 3 2) (h_diagonal1 node 2 2 2) (h_diagonal1 node 2 3 2) (h_diagonal2 node 3 1 2) (h_diagonal2 node 3 0 2) (h_diagonal2 node 2 1 2) (h_diagonal2 node 2 0 2))
      )
    )
  )
)


; Fa un recurregut per totes les fileres puntuant amb
; un 1 si troba 100, 010, 001 o amb un 10 si troba 110, 011, 101
; La funci noms mirar les poscicions x,x+1,x+2 de totes les fileres
; per tant s'ha de fer 2 crides amb x's diferents per mirar tot el tauler.
(defun h_y (node y x puntua)
  (if (= y 4)
    0
    (if (= (aref node y x) puntua)
      (if (= (aref node y (+ x 1)) puntua) ;1--
        (if (= (aref node y (+ x 2)) 0) ;11-
          (+ (h_y node (+ 1 y) x puntua) 10)  ; 110
          (h_y node (+ 1 y) x puntua)  ;112
        )
        (if (= (aref node y (+ x 1)) 0) ;1--
          (if (= (aref node y (+ x 2)) 0) ;10-
            (+ (h_y node (+ 1 y) x puntua) 1)  ; 100
            (if (= (aref node y (+ x 2)) puntua) ;10-
              (+ (h_y node (+ 1 y) x puntua) 10)  ; 101
              (h_y node (+ 1 y) x puntua)   ;102
            )
          )
          (h_y node (+ 1 y) x puntua)  ; 12-
        )
      )
      (if (= (aref node y x) 0)
        (if (= (aref node y (+ x 1)) puntua) ;0--
          (if (= (aref node y (+ x 2)) 0) ;01-
            (+ (h_y node (+ 1 y) x puntua) 1)  ; 010
            (if (= (aref node y (+ x 2)) puntua) ;01-
              (+ (h_y node (+ 1 y) x puntua) 10)  ; 011
              (h_y node (+ 1 y) x puntua) ;02-
            )
          )
          (if (= (aref node y (+ x 1)) 0) ;0--
            (if (= (aref node y (+ x 2)) puntua) ;00-
              (+ (h_y node (+ 1 y) x puntua) 1)  ; 001
              (h_y node (+ 1 y) x puntua)   ;000 o 002
            )
            (h_y node (+ 1 y) x puntua)   ;02-
          )
        )
        (h_y node (+ 1 y) x puntua)   ;2--
      )
    )
  )
)

; Fa un recurregut per totes les columnes puntuant amb
; un 1 si troba 100, 010, 001 o amb un 10 si troba 110, 011, 101
; La funci noms mirar les poscicions y,y+1,y+2 de totes les fileres
; per tant s'ha de fer 2 crides amb y's diferents per mirar tot el tauler.
(defun h_x (node y x puntua)
  (if (= x 4)
    0
    (if (= (aref node y x) puntua)
      (if (= (aref node (+ y 1) x) puntua) ;1--
        (if (= (aref node (+ y 2) x) 0) ;11-
          (+ (h_x node y (+ x 1) puntua) 10)  ; 110
          (h_x node y (+ x 1) puntua) ;112
        )
        (if (= (aref node (+ y 1) x) 0) ;1--
          (if (= (aref node (+ y 2) x) 0) ;10-
            (+ (h_x node y (+ x 1) puntua) 1)  ; 100
            (if (= (aref node (+ y 2) x) puntua) ;10-
              (+ (h_x node y (+ x 1) puntua) 10)  ; 101
              (h_x node y (+ x 1) puntua)   ;102
            )
          )
          (h_x node y (+ x 1) puntua)  ; 12-
        )
      )
      (if (= (aref node y x) 0)
        (if (= (aref node (+ y 1) x) puntua) ;0--
          (if (= (aref node (+ y 2) x) 0) ;01-
            (+ (h_x node y (+ x 1) puntua) 1)  ; 010
            (if (= (aref node (+ y 2) x) puntua) ;01-
              (+ (h_x node y (+ x 1) puntua) 10)  ; 011
              (h_x node y (+ x 1) puntua) ;02-
            )
          )
          (if (= (aref node (+ y 1) x) 0) ;0--
            (if (= (aref node (+ y 2) x) puntua) ;00-
              (+ (h_x node y (+ x 1) puntua) 1)  ; 001
              (h_x node y (+ x 1) puntua)   ;000 o 002
            )
            (h_x node y (+ x 1) puntua)   ;02-
          )
        )
        (h_x node y (+ x 1) puntua)   ;2--
      )
    )
  )
)

; Mira (y,x) (y-1,x-1) (y-2,x-2) puntuant amb
; un 1 si troba 100, 010, 001 o amb un 10 si troba 110, 011, 101
; Per cubrir totes les diagonals s'hauran de fer diverses crides amb
; x, y diferents.
(defun h_diagonal1 (node y x puntua)
      (if (= (aref node y x) puntua)
        (if (= (aref node (- y 1) (- x 1)) puntua)
          (if (= (aref node (- y 2) (- x 2)) 0)
            10 ; 110
            0 ;112
          )
          (if (= (aref node (- y 1) (- x 1)) 0)
            (if (= (aref node (- y 2) (- x 2)) puntua)
              10 ;101
              (if (= (aref node (- y 2) (- x 2)) 0)
                1 ;100
                0 ;102
              )
            )
            0 ;12-
          )
        )
        (if (= (aref node y x) 0)
          (if (= (aref node (- y 1) (- x 1)) puntua)
            (if (= (aref node (- y 2) (- x 2)) 0)
              1 ; 010
              (if (= (aref node (- y 2) (- x 2)) puntua)
                10 ; 011
                0 ;012
              )
            )
            (if (= (aref node (- y 1) (- x 1)) 0)
              (if (= (aref node (- y 2) (- x 2)) puntua)
                1 ; 001
                0 ;002 o 000
              )
              0 ;02-
            )
          )
          0 ;2--
        )
      )
)


; Mira (y,x) (y-1,x+1) (y-2,x+2) puntuant amb
; un 1 si troba 100, 010, 001 o amb un 10 si troba 110, 011, 101
; Per cubrir totes les diagonals s'hauran de fer diverses crides amb
; x, y diferents.
(defun h_diagonal2 (node y x puntua)
      (if (= (aref node y x) puntua)
        (if (= (aref node (- y 1) (+ x 1)) puntua)
          (if (= (aref node (- y 2) (+ x 2)) 0)
            10 ; 110
            0 ;112
          )
          (if (= (aref node (- y 1) (+ x 1)) 0)
            (if (= (aref node (- y 2) (+ x 2)) puntua)
              10 ;101
              (if (= (aref node (- y 2) (+ x 2)) 0)
                1 ;100
                0 ;102
              )
            )
            0 ;12-
          )
        )
        (if (= (aref node y x) 0)
          (if (= (aref node (- y 1) (+ x 1)) puntua)
            (if (= (aref node (- y 2) (+ x 2)) 0)
              1 ; 010
              (if (= (aref node (- y 2) (+ x 2)) puntua)
                10 ; 011
                0 ;012
              )
            )
            (if (= (aref node (- y 1) (+ x 1)) 0)
              (if (= (aref node (- y 2) (+ x 2)) puntua)
                1 ; 001
                0 ;002 o 000
              )
              0 ;02-
            )
          )
          0  ; 2--
        )
      )
)


;Retorna la columna a on l'ordinador considera millor ficar la seva ficha.
; * columnes numerades de 0 a 3 (d'esquerra a dreta)
(defun cami (node)
  (poda (copiar_tauler node) 0 -2000 2000) ; -2000 == -infinit, 2000 == infinit
)

;Algorisme poda-alfa-beta que retorna una posicio a on es millor ficar la ficha
; * Sabem si un nivell es max pq es par
;node ha de ser un tauler a on les posicions en blanc son 0, les fiches de
;l'ordinador sn 1 i les de l'usuari 2. El tauler es de dimensi 4x4 i les
;columnes estan numerades de 0 a 3 (d'esquerra a dreta).
(defun poda (node nivell alfa beta)
  (if (= nivell 16)  ; Aquest es el nivell de profunditat maxim
    (h node)
    (if (< alfa beta) ; Poda
      (let
        (
          ; Cada valor representa 1 dels 4 possibles nous estats
          (valor0 -2000)  ; Fiquem -2000 pq aixi en cas de no ser modificat
          (valor1 -2000)  ; a la funci (maxpos ...) aquest valor mai sera
          (valor2 -2000)  ; seleccionat
          (valor3 -2000)
        )
        ;;; Introduir ficha a posici 0 ;;;
        ; Comprovem que es pot ficar fitxa en aquesta posici
        (if (= (aref node 0 0) 0)
          (progn
            (setq valor0
              (if (evenp nivell)  ; Toca posar al pc
                    (if (guanya (ficha (copiar_tauler node) 0 1) 1) ; Si guanya
                      (- 1000 nivell)  ; No cal continuar expandint l'abre per aqui
                                       ; Com ms allunyat del nivell 0 pitjor pq
                                       ; triguem ms en guanyar
                      (poda (ficha (copiar_tauler node) 0 1) (+ nivell 1) alfa beta)
                    )
                    (if (guanya (ficha (copiar_tauler node) 0 2) 2) ; Si perd
                      (+ -1000 nivell) ; No cal continuar expandint l'abre per aqui
                                       ; Com ms allunyat del nivell 0 millor pq
                                       ; triguem ms en perdre
                      (poda (ficha (copiar_tauler node) 0 2) (+ nivell 1) alfa beta)
                    )
              )
            )
            (if (evenp nivell)  ; Si nivell max
              (if (> valor0 alfa)
                (setq alfa valor0)
                nil
              )
              (if (< valor0 beta) ; Si nivell min
                (setq beta valor0)
                nil
              )
            )
          )
        )
        ;;; Introduir ficha a posici 1 ;;;
        ; Comprovem que es pot ficar fitxa en aquesta posici
        (if (= (aref node 0 1) 0)
          (progn
            (setq valor1
              (if (evenp nivell)  ; Toca posar al pc
                    (if (guanya (ficha (copiar_tauler node) 1 1) 1) ; Si guanya
                      (- 1000 nivell) ; No cal continuar expandint l'abre per aqui
                                      ; Com ms allunyat del nivell 0 pitjor pq
                                      ; triguem ms en guanyar
                      (poda (ficha (copiar_tauler node) 1 1) (+ nivell 1) alfa beta)
                    )
                    (if (guanya (ficha (copiar_tauler node) 1 2) 2) ; Si perd
                      (+ -1000 nivell) ; No cal continuar expandint l'abre per aqui
                                       ; Com ms allunyat del nivell 0 millor pq
                                       ; triguem ms en perdre
                      (poda (ficha (copiar_tauler node) 1 2) (+ nivell 1) alfa beta)
                    )
              )
            )
            (if (evenp nivell)  ; Si nivell max
              (if (> valor1 alfa)
                (setq alfa valor1)
                nil
              )
              (if (< valor1 beta) ; Si nivell min
                (setq beta valor1)
                nil
              )
            )
          )
        )
        ;;; Introduir ficha a posici 2 ;;;
        ; Comprovem que es pot ficar fitxa en aquesta posici
        (if (= (aref node 0 2) 0)
          (progn
            (setq valor2
              (if (evenp nivell)  ; Toca posar al pc
                    (if (guanya (ficha (copiar_tauler node) 2 1) 1) ; Si guanya
                      (- 1000 nivell) ; No cal continuar expandint l'abre per aqui
                                      ; Com ms allunyat del nivell 0 pitjor pq
                                      ; triguem ms en guanyar
                      (poda (ficha (copiar_tauler node) 2 1) (+ nivell 1) alfa beta)
                    )
                    (if (guanya (ficha (copiar_tauler node) 2 2) 2) ; Si perd
                      (+ -1000 nivell) ; No cal continuar expandint l'abre per aqui
                                       ; Com ms allunyat del nivell 0 millor pq
                                       ; triguem ms en perdre
                      (poda (ficha (copiar_tauler node) 2 2) (+ nivell 1) alfa beta)
                    )
              )
            )
            (if (evenp nivell)  ; Si nivell max
              (if (> valor2 alfa)
                (setq alfa valor2)
                nil
              )
              (if (< valor2 beta) ; Si nivell min
                (setq beta valor2)
                nil
              )
            )
          )
        )
        ;;; Introduir ficha a posici 3 ;;;
        ; Comprovem que es pot ficar fitxa en aquesta posici
        (if (= (aref node 0 3) 0)
          (progn
            (setq valor3
              (if (evenp nivell)  ; Toca posar al pc
                    (if (guanya (ficha (copiar_tauler node) 3 1) 1) ; Si guanya
                      (- 1000 nivell) ; No cal continuar expandint l'abre per aqui
                                      ; Com ms allunyat del nivell 0 pitjor pq
                                      ; triguem ms en guanyar
                      (poda (ficha (copiar_tauler node) 3 1) (+ nivell 1) alfa beta)
                    )
                    (if (guanya (ficha (copiar_tauler node) 3 2) 2) ; Si perd
                      (+ -1000 nivell) ; No cal continuar expandint l'abre per aqui
                                       ; Com ms allunyat del nivell 0 millor pq
                                       ; triguem ms en perdre
                      (poda (ficha (copiar_tauler node) 3 2) (+ nivell 1) alfa beta)
                    )
              )
            )
            (if (evenp nivell)  ; Si nivell max
              (if (> valor3 alfa)
                (setq alfa valor3)
                nil
              )
              (if (< valor3 beta) ; Si nivell min
                (setq beta valor3)
                nil
              )
            )
          )
        )
        ; Si es l'ultim nivell, ja hem acabat
        (if (= nivell 0)
          ; Retornem la posici del valor maxim, la qual es correspondra amb
          ; la posici al tauler a on ens convindra ficar la ficha.
          (maxpos (list valor0 valor1 valor2 valor3))
          (if (evenp nivell) 
            alfa
            beta
          )
        )
      )
      ;Aixo s'executa si es fa la poda
      (if (evenp nivell)
        alfa
        beta
      )
    )
  )
)

; Jugar en un tauler comenant l'ordinador
(defun jugar1 (node)
  (if (guanya node 2)
    (print "Guanya l'usuari")
    (if (ple node)
      (print "Empat")
      (let
        (
          (pos (cami node))
        )
        (print "L'ordinador fica la seva ficha a columna nmero...")
        (print pos)
        (print (ficha node pos 1))
        (if (guanya node 1)
          (print "Guanya l'ordinador")
          (progn
            (print "A quina columna vol ficar la ficha?")
            (jugar1 (ficha node (read) 2))
          )
        )
      )
    )
  )
)

; Jugar en un tauler comenant l'usuari
(defun jugar2 (node)
  (if (guanya node 1)
    (print "Guanya l'ordinador")
    (if (ple node)
      (print "Empat")
      (let
        (
          (pos nil)
        )
        (print "A quina columna vol ficar la ficha?")
        (ficha node (read) 2)
        (if (guanya node 2)
          (print "Guanya l'usuari")
          (progn
            (setq pos (cami node))
            (print "L'ordinador fica la seva ficha a columna nmero...")
            (print pos)
            (print (ficha node pos 1))
            (jugar2 node)
          )
        )
      )
    )
  )
)

;Juga contra ell mateix
(defun jugar3 (node)
  (if (guanya node 2)
    (print "Guanya l'ordinador2")
    (if (ple node)
      (print "Empat")
      (let
        (
          (pos1 (cami node))
          (pos2 nil)
        )
        (print "L'ordinador1 fica la seva ficha a columna nmero...")
        (print pos1)
        (print (ficha node pos1 1))
        (if (guanya node 1)
          (print "Guanya l'ordinador1")
          (progn
            (print node)
            (setq pos2 (cami node))
            (print "L'ordinador2 fica la seva ficha a columna nmero...")
            (print pos2)
            (jugar3 (ficha node pos2 2))
          )
        )
      )
    )
  )
)

;Funcio principal per jugar. Et permet escullir si volem ser nosaltres els primers
;en llenar ficha o no.
(defun jugar ()
  (let
    (
      (tauler (make-array '(4 4) :initial-contents '((0 0 0 0)(0 0 0 0)(0 0 0 0)(0 0 0 0))))
      (answer nil)
    )
    (print " 1.- Ordinador")
    (print " 2.- Jugador")
    (print " 3.- Ordinador VS Ordinador")
    (print "Qui vol que inicii la partida?")
    (setq answer (read))
    (if (= answer 2)
      (jugar2 tauler)
      (if (= answer 3)
        (jugar3 tauler)
        (jugar1 tauler)
      )
    )
  )
)

